c
c     airshed.f - Airshed simulation
c
c     CMU task parallel program suite
c
c     Peter Dinda, Thomas Gross, David O'Hallaron, Edward Segall,
c     James Stichnoth, Jaspal Subhlok, Jon Webb, Bwolen Yang
c
c     Adapted from the airshed model introduced by G. McRae, W. Goodin, 
c     and J. Seinfeld, "Development of a second-generation mathematical
c     model for urban air pollution - 1. Model formulation", Atmospheric
c     Environment, v 16, n 4, pp. 679-696, 1982, and implemented by
c     G. McRae, A. Russell, and R. Harley, "CIT photochemical airshed
c     model - systems manual", Carnegie Mellon University, Pittsburgh,
c     PA, and California Institute of Technology, Pasadena, CA., 1992.
c
c     Comments and suggestions to David O'Hallaron, droh@cs.cmu.edu


cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c     Copyright (c) 1994 by Carnegie Mellon University
c
c     Permission to use, copy, modify, and distribute this software
c     for any purpose and without fee is hereby granted, provided that 
c     the above copyright notice appear in all copies.
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      program airshed
      include "airshed.h"

c     the  major data structures
      real C(NLAYERS,NSPECIES,NCOLS)    ! concentration matrix
      real U(NLAYERS,NCOLS)             ! wind velocities 
      real V(NLAYERS,NCOLS)  
      real KK(NLAYERS,BW,NCOLS)         ! stiffness matrices

c     miscellaneous
      integer hour, nsteps, istep
      
c     timing variables
      real isecs, ssecs, tsecs, csecs, seconds, start 

c     begin the airshed simulation
      isecs = 0
      ssecs = 0
      tsecs = 0
      csecs = 0
      call init(C)
      do hour = 1, NHRS
         start = seconds()
         call inputhour(C, U, V, nsteps)    
         isecs = isecs + seconds()
         call pretrans(KK, U, V)
         ssecs = ssecs + seconds()
         do istep = 1, nsteps
            call transport(C, KK, U, V)
            tsecs = tsecs + seconds()
            call chemistry(C)
            csecs = csecs + seconds()
            call transport(C, KK, U, V)
            tsecs = tsecs + seconds()
         enddo
         call chkmat(C,hour)
      enddo
      call prperf(isecs, ssecs, tsecs, csecs)
      end

c
c     init - get the initial concentration matrix
c            OUT C
c
      subroutine init(C)
      include "airshed.h"
      real C(NLAYERS,NSPECIES,NCOLS)
      real val
      integer i, j, k

      val = 1.0
      do i = 1, NLAYERS
         do j = 1, NSPECIES
            do k = 1, NCOLS
               C(i,j,k) = val
               val = val + 1.0
            enddo
         enddo
      enddo
      end

c
c     inputhour - read this hour's data 
c                 INOUT C; OUT U, V, nsteps
c
      subroutine inputhour(C, U, V, nsteps)
      include "airshed.h"
      real C(NLAYERS,NSPECIES,NCOLS)
      real U(NLAYERS,NCOLS)
      real V(NLAYERS,NCOLS)
      real B(NSPECIES,NCOLS)
      integer nsteps
      real fdelay
      integer i, j, k

c     input the number of steps
      nsteps = DEFNSTEPS

c     update the first layer of the concentration matrix
      do j = 1, NSPECIES
         do k = 1, NCOLS
            B(j,k) = fdelay(DINPUT)
         enddo
      enddo
      do j = 1, NSPECIES
         do k = 1, NCOLS
            C(1,j,k) = C(1,j,k) + B(j,k)
         enddo
      enddo

c     input the velocity vectors
      do i = 1, NLAYERS
         do k = 1, NCOLS
            U(i,k) = fdelay(DINPUT)
            V(i,k) = fdelay(DINPUT)
         enddo
      enddo
      end

c
c     pretrans -  proprocessing for the next hour's transport steps     
c                  IN U, V; OUT KK
c
      subroutine pretrans(KK, U, V)
      include "airshed.h"
      real KK(NLAYERS,NCOLS,NCOLS)
      real U(NLAYERS,NCOLS)
      real V(NLAYERS,NCOLS)
      integer ilayer

c     assemble and factor the global stiffness matrix for each layer
      do ilayer = 1, NLAYERS
         call assemble1(KK, U, V, ilayer)
         call factor1(KK, ilayer)
      enddo
      end

      
c
c     transport - FEM simulation of the movement of airborne pollutants
c                 IN KK, U, V; INOUT C
c     
      subroutine transport(C, KK, U, V)
      include "airshed.h"
      real C(NLAYERS,NSPECIES,NCOLS)
      real KK(NLAYERS,BW,NCOLS)
      real U(NLAYERS,NCOLS)      
      real V(NLAYERS,NCOLS)  
      integer ilayer, ispec

      do ilayer = 1,NLAYERS
         do ispec = 1,NSPECIES
            call trans1(C, KK, ilayer, ispec)
            call filter1(C, U, V, ilayer, ispec)
         enddo
      enddo
      end

c
c     chemistry - chemistry computation on independent air columns
c                 INOUT C      
c
      subroutine chemistry(C)
      include "airshed.h"
      real C(NLAYERS,NSPECIES,NCOLS)
      integer icol

      do icol = 1, NCOLS
         call chem1(C, icol)
      enddo
      end

c
c     Sequential routines
c
c     The next five routines: assemble1(), factor1(), trans1(), 
c     filter1(), and chem1(), represent synthetic workloads and should 
c     be run sequentially. They shouldn't be parallelized.
c

c
c     assemble1 - assemble the stiffness matrix for layer ilayer
c                IN U(ilayer,:), V(ilayer,:), ilayer; 
c                OUT KK(ilayer,:,:)
c
      subroutine assemble1(KK, U, V, ilayer)
      include "airshed.h"
      real KK(NLAYERS,BW,NCOLS)
      real U(NLAYERS,NCOLS)
      real V(NLAYERS,NCOLS)
      real fdelay
      integer ilayer, j, k
      
      do j = 1, BW
         do k = 1, NCOLS
            KK(ilayer,j,k) = U(ilayer,k) + V(ilayer,k) + 
     $           fdelay(DASSEMBLE) 
         enddo
      enddo
      end

c
c     factor1 - factor the stiffness matrix for layer ilayer
c              IN ilayer, INOUT KK(ilayer,:,:)
c
      subroutine factor1(KK, ilayer)
      include "airshed.h"
      real KK(NLAYERS,BW,NCOLS)
      real fdelay
      integer ilayer, j, k

      do j = 1, BW
         do k = 1, NCOLS
               KK(ilayer,j,k) = KK(ilayer,j,k) + fdelay(BW*DFACTOR)
         enddo
      enddo
      end

c
c     trans1 - backsolve for one layer/species of C
c              IN KK(ilayer,:,:), ilayer, ispec; INOUT C(ilayer,ispec,:)
c
      subroutine trans1(C, KK, ilayer, ispec)
      include "airshed.h"
      real C(NLAYERS,NSPECIES,NCOLS)
      real KK(NLAYERS,BW,NCOLS) 
      real val, fdelay
      integer ilayer, ispec, icol, k

      do icol = 1, NCOLS
         val = 0.0
         do k=1,BW
            val = val + KK(ilayer,k,icol)
         enddo
         C(ilayer,ispec,icol) = C(ilayer,ispec,icol) + 
     $        val + fdelay(DTRANSPORT)
      enddo
      end

c
c     filter1 - filter one layer/species C using U and V
c               IN U(ilayer,:), V(ilayer,;), ilayer, ispec; 
c               INOUT C(ilayer,ispec,:)
c
      subroutine filter1(C, U, V, ilayer, ispec)
      include "airshed.h"
      real C(NLAYERS,NSPECIES,NCOLS)
      real U(NLAYERS,NCOLS)      
      real V(NLAYERS,NCOLS)  
      real fdelay
      integer ilayer, ispec, icol

      do icol = 1, NCOLS
         C(ilayer,ispec,icol) = C(ilayer,ispec,icol) + 
     $        U(ilayer,icol) + V(ilayer,icol) + fdelay(DFILTER)
      enddo
      end

c
c     chem1 - computation for one independent column
c             IN icol; INOUT C(:,:,icol)
c
      subroutine chem1(C, icol)
      include "airshed.h"
      real C(NLAYERS,NSPECIES,NCOLS)
      real fdelay
      integer icol, ilayer, ispec

      do ilayer = 1, NLAYERS
         do ispec = 1, NSPECIES
            C(ilayer,ispec,icol) = 
     $           C(ilayer,ispec,icol) + fdelay(DCHEMISTRY)
         enddo
      enddo
      end


c
c     Support routines
c

c
c     chkmat - check the concentration matrix for "correct" results
c
      subroutine chkmat(C,hour)
      include "airshed.h"
      real C(NLAYERS,NSPECIES,NCOLS)
      integer ilayer, ispec, icol, errors, hour
      real val, ref, kkval, transval

      kkval = 2*DINPUT + DASSEMBLE + BW*DFACTOR 
      transval = BW*kkval + DTRANSPORT + 2*DINPUT + DFILTER
      errors = 0
      val = 1.0
      do ilayer = 1,NLAYERS
         do ispec = 1,NSPECIES
            do icol = 1,NCOLS
               ref = val + hour*(DEFNSTEPS*((2*transval) + DCHEMISTRY))
               if (ilayer .eq. 1) then
                  ref = ref + hour*DINPUT        
               endif
               if (C(ilayer,ispec,icol) .ne. ref) then
                  errors = errors+1
                  print *,ilayer,ispec,icol,C(ilayer,ispec,icol),ref
               endif
               val = val + 1.0
            enddo
         enddo
      enddo
      if (errors .gt. 0) then
         print *, errors, " errors detected in hour ", hour
      endif
      end

c
c     prperf - summarize the performance
c
      subroutine prperf(isecs, ssecs, tsecs, csecs)
      include "airshed.h"
      real isecs, ssecs, tsecs, csecs, secs
      integer ipercent, spercent, tpercent, cpercent

      isecs = isecs/real(NHRS)
      ssecs = ssecs/real(NHRS)
      tsecs = tsecs/real(NHRS)
      csecs = csecs/real(NHRS)
      secs = isecs + ssecs + tsecs + csecs

      ipercent = (isecs/secs)*100
      spercent = (ssecs/secs)*100
      tpercent = (tsecs/secs)*100
      cpercent = (csecs/secs)*100

      print '(''************************************'')'
      print '(''Airshed simulation'')'
      print '(i1,'' layers, '',i2,'' species, '',i4,'' columns'')',
     $     NLAYERS, NSPECIES, NCOLS
      print '(''************************************'')'
      print '(''Time per simulated hour:'')'
      print *, " "
      print '(''inputhour: '',f9.2,'' ('',i2,''%)'')',isecs,ipercent
      print '(''setuphour: '',f9.2,'' ('',i2,''%)'')',ssecs,spercent
      print '(''transport: '',f9.2,'' ('',i2,''%)'')',tsecs,tpercent
      print '(''chemistry: '',f9.2,'' ('',i2,''%)'')',csecs,cpercent
      print '(''total(s) : '',f9.2)',secs
      print *, " "
      end

c
c     seconds - returns elapsed seconds since last call to seconds().
c               the first time, returns elapsed time since program started. 
c
      real function seconds()
      real tarray(2),dtime
      seconds = dtime(tarray)
      seconds = tarray(1)
      end

c
c     ndelay - delay for n iterations and then return n
c
c     Note: The code is a little involved because, while we want the loop 
c     to just delay n iterations and then return n, we don't want a smart 
c     optimizer to eliminate the loop. The mod function in the loop body 
c     is designed to stop the optimizer from recognizing the reduction.
c     The use of val in the last conditional forces the compiler to 
c     execute the delay loop. 
c
      real function fdelay(n)
      integer n, i
      real val

      val = 1.0
      do i = 1,n
         if (mod(i,2) .eq. 0) then
            val = val + 1.0
         endif
      enddo
      if (val .le. n) then   ! always true (but the compiler doesn't know!)
         fdelay = real(n)
      else                   ! always false
         print *, "system error: bad delay logic"
         stop
      endif
      end



