c
c     fft1.f - 1D fast Fourier transform
c
c     CMU task parallel program suite
c
c     Peter Dinda, Thomas Gross, David O'Hallaron, Edward Segall,
c     James Stichnoth, Jaspal Subhlok, Jon Webb, Bwolen Yang
c
c     Bailey's Six-Step Method, adapted from D. Bailey, 
c     "FFTs in External or Hierarchical Memory", 
c     The Journal of Supercomputing, 4, 23-35 (1990).
c
c     Comments and suggestions to David O'Hallaron, droh@cs.cmu.edu

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c     Copyright (c) 1994 by Carnegie Mellon University
c
c     Permission to use, copy, modify, and distribute this software
c     for any purpose and without fee is hereby granted, provided that 
c     the above copyright notice appear in all copies.
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      program fft1
      include "fft1.h"

      real xin(2,N,N)         ! input vector reshaped as an array
      real a(2,N,N)           ! intermediate arrays
      real b(2,N,N)

c     precomputed constants
      integer brt(N)         ! bit reverse table for 1d fft's
      real w(2,N/2)          ! twiddles for 1d fft's
      real v(2,N,N)          ! twiddles for 2d fft

c     timing variables
      real tsecs,ssecs,fsecs 
      real seconds,start

c     misc
      integer k
c     
c     precompute the input array and fft constants
c     
      call gen_bit_reverse_table(brt,N)
      call gen_w_table(w,N,LOGN,N/2)
      call gen_v_table(v,N)
c
c     begin the 1d fft program 
c
      tsecs = 0
      ssecs = 0
      fsecs = 0
      do k=1,ITERS
         call dgen(xin,N)      
         start = seconds()                ! restart the timer
         call tpose(xin,b,N)              ! step 1
         tsecs = tsecs + seconds()
         call cffts(b,brt,w,N,LOGN,N/2)   ! step 2
         fsecs = fsecs + seconds()
         call scale(b,v,N)                ! step 3
         ssecs = ssecs + seconds()
         call tpose(b,a,N)                ! step 4
         tsecs = tsecs + seconds()
         call cffts(a,brt,w,N,LOGN,N/2)   ! step 5
         fsecs = fsecs + seconds()
         call tpose(a,b,N)                ! step 6
         tsecs = tsecs + seconds()
         call chkmat(b,N)                 ! check the results
      enddo
      call prperf(tsecs,ssecs,fsecs,NN,LOGNN,ITERS)
      end
      
c     
c     dgen - initialize the vector with a point source
c     
      subroutine dgen(xin,n)
      integer n
      real xin(2,n,n)
      integer i,j
      
      do j = 1,n
         do i = 1,n
            xin(1,i,j) = 0.0
            xin(2,i,j) = 0.0
         enddo
      enddo
      xin(1,1,n/2+1) = n*n
      end


c
c     gen_bit_reverse_table - initialize bit reverse table
c       
c     Postcondition: br(i) = bit-reverse(i-1) + 1       
c
      subroutine gen_bit_reverse_table(brt,n)
      integer n
      integer brt(n)
      integer i, j, k, nv2

      nv2 = n/2
      j = 1
      brt(1) = j
      do i = 2, n
         k = nv2
 6       continue
         if (k .lt. j) then
            j = j - k
            k = k/2
            goto 6
         endif
         j = j + k
         brt(i) = j
      enddo
      return
      end

c
c     gen_w_table: generate powers of w.
c
c           postcondition: w(i) = w**(i-1)            
c
      subroutine gen_w_table(w,n,logn,ndv2)
      integer n,logn,ndv2
      real w(2,ndv2)
      integer i
      real pi
      real wr,wi
      real ptr, pti
      
      pi = 3.141592653589793

      wr = cos(pi/ndv2)
      wi = -sin(pi/ndv2)
      w(1,1) = 1.0
      w(2,1) = 0.0
      ptr = 1.0
      pti = 0.0
      do i = 2,ndv2
         w(1,i) = ptr*wr - pti*wi
         w(2,i) = ptr*wi + pti*wr
         ptr = w(1,i)
         pti = w(2,i)
      enddo
      return
      end

c
c     gen_v_table - gen 2d twiddle factors
c
      subroutine gen_v_table(v,n)
      integer n
      real v(2,n,n)
      complex wn,w
      real pi
      integer k,j
      
      pi = 3.141592653589793
      wn = cmplx(cos((2.0*pi)/(n*n)),-sin((2.0*pi)/(n*n)))
      do k = 1,n
         do j = 1,n
            w = wn**((j-1)*(k-1))
            v(1,j,k) = real(w)
            v(2,j,k) = imag(w)
         enddo
      enddo
      end

c
c     scale - multiply each array element by corresponding element of v
c      
      subroutine scale(a,v,n)
      integer n
      real a(2,n,n),v(2,n,n)

      real ar,ai,vr,vi
      integer i,j

      do j = 1,n
         do i = 1,n
            ar = a(1,i,j)
            ai = a(2,i,j)
            vr = v(1,i,j)
            vi = v(2,i,j)
            a(1,i,j) = ar*vr - ai*vi
            a(2,i,j) = ar*vi + ai*vr
         enddo
      enddo
      end

c
c     tpose - matrix transpose
c
      subroutine tpose(a,b,n)
      integer n,i,j
      real a(2,n,n),b(2,n,n),tr,ti

      do i = 1,n
         do j = i,n
            tr = a(1,i,j)
            ti = a(2,i,j)
            b(1,i,j) = a(1,j,i)
            b(2,i,j) = a(2,j,i)
            b(1,j,i) = tr
            b(2,j,i) = ti
         enddo
      enddo
      end

c     
c     cffts - perform a 1d fft on each column 
c     
      subroutine cffts(a,brt,W,n,logn,ndv2)
      integer n,logn,ndv2,brt(n)
      real a(2,n,n),w(2,ndv2)
      integer col

      do col=1,n
         call fft(a(1,1,col),brt,w,n,logn,ndv2)
      enddo
      end
      
c
c     Fast Fourier Transform 
c     1D in-place complex-complex decimation-in-time Cooley-Tukey
c
      subroutine fft(a, brt, W, n, logn, ndv2)
      integer n,logn,ndv2
      real a(2,n)
      integer brt(n)
      real W(2,ndv2)

      integer i,j
      integer powerOfW
      integer sPowerOfW
      integer ijDiff
      integer stage
      integer stride
      integer first
      real pwr,pwi,tr,ti
      real ir, ii, jr, ji
c     
c     bit reverse step
c     
      do i = 1,n
         j = brt(i)
         if (i .lt. j) then
            tr = a(1,j)
            ti = a(2,j)
            a(1,j) = a(1,i)
            a(2,j) = a(2,i)
            a(1,i) = tr
            a(2,i) = ti
         endif
      enddo
         
c     
c     butterfly computations
c     
      ijDiff = 1
      stride = 2
      sPowerOfW = ndv2
      do stage = 1,logn
c     Invariant: stride = 2 ** stage            
c     Invariant: ijDiff = 2 ** (stage - 1)
         
         first = 1
         do powerOfW = 1, ndv2, sPowerOfW
            pwr = W(1,powerOfW)
            pwi = W(2,powerOfW)
            
c     Invariant: pwr + sqrt(-1)*pwi = W**(powerOfW - 1)

            do i = first, n, stride
               j = i + ijDiff
               jr = a(1,j)
               ji = a(2,j)
               ir = a(1,i)
               ii = a(2,i)
               tr = jr*pwr - ji*pwi
               ti = jr*pwi + ji*pwr
               a(1,j) = ir - tr
               a(2,j) = ii - ti
               a(1,i) = ir + tr
               a(2,i) = ii + ti
            enddo

            first = first + 1
         enddo
            
         ijDiff = stride
         stride = stride * 2
         sPowerOfW = sPowerOfW / 2
      enddo
      return
      end
 
c     
c     chkmat - check the output matrix for correctness
c     
      subroutine chkmat(a,n)
      integer n
      real a(2,n,n)

      integer i,j,nn
      integer sign,errors
      real epsilon

      epsilon = 0.0001
      errors = 0
      nn = n*n
      do j=1,n
         sign = 1
         do i=1,n
            if (a(1,i,j) .gt. nn*sign+epsilon) errors = errors + 1
            if (a(1,i,j) .lt. nn*sign-epsilon) errors = errors + 1
            sign = sign * (-1)
         enddo
      enddo
      if (errors .gt. 0) then
         print *, 'errors = ', errors
         stop
      endif
      end
      
c     
c     output routines
c     
      subroutine prperf(tsecs,ssecs,fsecs,nn,lognn,iters)
      integer nn,lognn,iters
      real tsecs,ssecs,fsecs,secs
      real flops,mflops
      integer tpercent,spercent,fpercent
      
      tsecs = tsecs/iters
      ssecs = ssecs/iters
      fsecs = fsecs/iters
      secs = tsecs+ssecs+fsecs

      tpercent = (tsecs/secs)*100
      spercent = (ssecs/secs)*100
      fpercent = (fsecs/secs)*100

      flops = 5*nn*lognn
      mflops = (flops/1000000.0)/secs
      print '(''***********************'')'
      print '(''1D FFT ('',i6,'' points)'')',nn
      print '(''***********************'')'
      print '(''Time per input vector:'')'
      print '(''tpose    : '',f9.6,'' ('',i2,''%)'')',tsecs,tpercent
      print '(''scale    : '',f9.6,'' ('',i2,''%)'')',ssecs,spercent
      print '(''ffts     : '',f9.6,'' ('',i2,''%)'')',fsecs,fpercent
      print '(''total(s) : '',f9.6)',secs
      print *," "
      print '(''mflop/s  : '',f7.2)', mflops
      print *, " "
      end

      subroutine print_cmat(a,n)
      integer i,j,n
      real a(2,n,n)

      do i=1,n
         print '(8 (:''(''f6.2,f6.2'')'' 1x))',(a(1,i,j),a(2,i,j),j=1,n)
      enddo
      end

      subroutine print_cvec(v, n)
      integer n, i
      real v(2,n)

      do i=1,n
         print '(''(''f6.2,f6.2'')'')', v(1,i), v(2,i)
      enddo
      end

      subroutine print_ivec(v, n)
      integer n, v(n), i

      do i=1,n
         print *, v(i)
      enddo
      end

c
c     seconds - returns elapsed seconds since last call to seconds().
c               the first time, returns elapsed time since program started. 
c
      real function seconds()
      real tarray(2),dtime
      seconds = dtime(tarray)
      seconds = tarray(1)
      end
