c 
c     radar.f - Narrowband tracking radar
c 
c     CMU task parallel program suite
c
c     Peter Dinda, Thomas Gross, David O'Hallaron, Edward Segall,
c     James Stichnoth, Jaspal Subhlok, Jon Webb, Bwolen Yang
c
c     Adapted from G. Shaw et al, "Multiprocessors for Radar Signal 
c     Processing", Lincoln Laboratories, MIT, Lexington, MA, Tech 
c     Report 961, 17 Nov 1992.
c
c     Comments and suggestions to David O'Hallaron, droh@cs.cmu.edu

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c     Copyright (c) 1994 by Carnegie Mellon University
c
c     Permission to use, copy, modify, and distribute this software
c     for any purpose and without fee is hereby granted, provided that 
c     the above copyright notice appear in all copies.
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      program radar
      include 'radar.h'

c     major data sets 
      real inpr(2,NRANGE,NDOP,NCHAN)  ! input (before corner turn)
      real inp(2,NDOP,NRANGE,NCHAN)   ! input (after corner turn)
      integer det(NDOP,NRANGE,NCHAN)  ! output detect array

c     intermediate data sets 
      real w(NDOP)                ! hamming window 

c     precomputed FFT data structures 
      integer brt(NDOP)        ! bit reverse table
      real twiddles(2,NDOP/2)  ! coefficients

c     target ranges and frequencies
      integer range0(NTARGETS)    
      integer doppler0(NTARGETS)  

c     misc items
      integer k,iter
      real sigma0,thresh,const2,const1,pi,a0

c     timing variables 
      real fsecs,tsecs,dsecs,start,seconds
      common /timing/ fsecs,tsecs,dsecs

c     initialize constants
      pi = 3.141592653589793
      sigma0 = 1.13  !0.0->noiseless; 1.13->20dB SNR; 3.58->10dB SNR 
      thresh = 2.421
      const2 = thresh/(NRANGE*(K1-K0+1)-1+thresh)
      const1 = 2*pi/NDOP
      
c     target frequencies
      doppler0(1) = 6
      doppler0(2) = 16
      doppler0(3) = 26
      doppler0(4) = 36

c     target ranges
      range0(1) = 3
      range0(2) = 5
      range0(3) = 7
      range0(4) = 9

c     onetime setup for FFT
      call gen_bit_reverse_table(brt,NDOP)
      call gen_W_table(twiddles,NDOP,MDOP,NDOP/2)

c     precompute the hamming window
      a0 = 0.5*(NDOP-1)
      do k = 1,NDOP
         w(k) = 0.54 + 0.46*cos(const1*(k-1-a0))
      enddo

c
c     The radar program 
c
      tsecs = 0 ! transpose time
      fsecs = 0 ! fft/mag time
      dsecs = 0 ! detect time
      do iter = 1,ITERS
         call dgen(inpr, range0, doppler0, sigma0)
         start = seconds()
         call corner_turn(inpr,inp)
         call compute(inp, det, brt, twiddles, w, const2)
      enddo
      call chkmat(det)
      call prperf(tsecs,fsecs,dsecs,ITERS)
      end

c
c     corner_turn - change array in range vector order to
c                   to one that is in doppler vector order.
c
      subroutine corner_turn(inpr,inp)
      include 'radar.h'
      real inpr(2,NRANGE,NDOP,NCHAN)
      real inp(2,NDOP,NRANGE,NCHAN)
      integer i,j,k
      
      real fsecs,tsecs,dsecs,seconds
      common /timing/ fsecs,tsecs,dsecs

      do k = 1,NCHAN
         do j = 1,NRANGE
            do i = 1,NDOP
               inp(1,i,j,k) = inpr(1,j,i,k)
               inp(2,i,j,k) = inpr(2,j,i,k)
            enddo
         enddo
      enddo
      tsecs = tsecs + seconds()
      end

c
c     compute - process the corner turned input samples
c
      subroutine compute(inp, det, brt, twiddles, w, const2)
      include 'radar.h'
      real inp(2,NDOP,NRANGE,NCHAN)   ! sensor inputs (before corner turn)
      integer det(NDOP,NRANGE,NCHAN)  ! detect array
      integer brt(NDOP)               ! bit reverse table
      real twiddles(2,NDOP/2)         ! fft coefficients
      real w(NDOP)                    ! hamming window 
      real const2                     ! precomputed threshhold constant
      
c     intermediate arrays
      real x(2,NDOP)              ! windowed FFT input 
      real y(2,NDOP)              ! scaled  FFT output 
      real localthresh(NCHAN)     ! threshholds 
      real mag(NDOP,NRANGE,NCHAN) ! magnitude array
      real magsum
      integer i,j,k

      real fsecs,tsecs,dsecs,seconds
      common /timing/ fsecs,tsecs,dsecs

      do i = 1, NCHAN
c     windowing and Doppler processing
         do j = 1,NRANGE
            do k = 1,NDOP
               x(1,k) = inp(1,k,j,i)*w(k)
               x(2,k) = inp(2,k,j,i)*w(k)
            enddo
            call fft(x, brt, twiddles, NDOP, MDOP, NDOP/2)
            do k = K0,K1
               y(1,k) = x(1,k)/NDOP
               y(2,k) = x(2,k)/NDOP
            enddo
            do k = K0,K1
               mag(k,j,i) = sqrt(y(1,k)*y(1,k) + y(2,k)*y(2,k))
            enddo
         enddo
         fsecs = fsecs + seconds()
         
c     Doppler subselection and reduction
         magsum = 0.0
         do j = 1,NRANGE
            do k = K0,K1
               magsum = magsum + mag(k,j,i)
            enddo
         enddo
         
c     constant false alarm rate (CFAR) detection 
         localthresh(i) = const2 * magsum
         do j = 1,NRANGE
            do k = K0,K1
               if (mag(k,j,i) .gt. localthresh(i)) then
                  det(k,j,i) = 1
               else
                  det(k,j,i) = 0
               endif
            enddo
         enddo
         dsecs = dsecs + seconds()
      enddo !i loop
      end

c     
c     dgen - generate the input data set 
c
      subroutine dgen(inpr, range0, doppler0, sigma0)
      include 'radar.h'
      
      real inpr(2,NRANGE,NDOP,NCHAN)
      integer range0(NTARGETS), doppler0(NTARGETS)
      real sigma0
      integer i, j, k, m, range, doppler
      real arg1, arg2, pi
      real gauss1
      
      do i = 1,NCHAN
         do k = 1,NDOP
            do j = 1,NRANGE
               inpr(1,j,k,i) = gauss1(0.0,sigma0)
               inpr(2,j,k,i) = gauss1(0.0,sigma0)
            enddo
         enddo
      enddo

      pi = 3.141592653589793
      arg1 = 2.0*pi/NDOP
      do m = 1, NTARGETS
         range = range0(m)
         doppler = doppler0(m)
         do i = 1, NCHAN
            do k = 1,NDOP
               arg2 = arg1*(k-1)*doppler
               inpr(1,range,k,i) = inpr(1,range,k,i) + cos(arg2)
               inpr(2,range,k,i) = inpr(2,range,k,i) + sin(arg2)
            enddo
         enddo
      enddo
      end
c
c     gauss1 - crude approximation of normal distribution formed
c              from sum of 12 uniforms.
c
      real function gauss1(mean,sigma)
      real mean, sigma
      real random1
      
      integer i
      real x

      x = 0.0
      do i=1,12
         x = x + real(random1())
      enddo
      gauss1 = mean + sigma*(x - 6.0)
      end

c 
c random1 - uniformly distributed [0,1) real
c
      real function random1() 
      save a
      data a/11111111/
      integer a, m, m1, b, a0, a1, b0, b1

      m = 100000000
      m1 = 10000
      b = 31415821

      a1 = a/m1
      a0 = mod(a,m1)
      b1 = b/m1
      b0 = mod(b,m1)
      a0 = mod((a0*b1 + a1*b0),m1)*m1 + a0*b0
      a = mod((a0 + 1),m)
      random1 = real(a)/m
      end


c
c     gen_bit_reverse_table - initialize bit reverse table 
c       
c     Postcondition: br(i) = bit-reverse(i-1) + 1       
c
      subroutine gen_bit_reverse_table(brt,n)
      integer n
      integer brt(n)
      integer i, j, k, nv2

      nv2 = n/2
      j = 1
      brt(1) = j
      do i = 2, n
         k = nv2
 6       continue
         if (k .lt. j) then
            j = j - k
            k = k/2
            goto 6
         endif
         j = j + k
         brt(i) = j
      enddo
      return
      end

c
c     gen_w_table: generate powers of w.
c
c           postcondition: w(i) = w**(i-1)            
c
      subroutine gen_w_table(w,n,logn,ndv2)
      integer n,logn,ndv2
      real w(2,ndv2)
      integer i
      real pi
      real wr,wi
      real ptr, pti
      
      pi = 3.141592653589793

      wr = cos(pi/ndv2)
      wi = -sin(pi/ndv2)
      w(1,1) = 1.0
      w(2,1) = 0.0
      ptr = 1.0
      pti = 0.0
      do i = 2,ndv2
         w(1,i) = ptr*wr - pti*wi
         w(2,i) = ptr*wi + pti*wr
         ptr = w(1,i)
         pti = w(2,i)
      enddo
      return
      end

c
c     Fast Fourier Transform 
c     1D in-place complex-complex decimation-in-time Cooley-Tukey
c
      subroutine fft(a, brt, W, n, logn, ndv2)
      integer n,logn,ndv2
      real a(2,n)
      integer brt(n)
      real W(2,ndv2)

      integer i,j
      integer powerOfW
      integer sPowerOfW
      integer ijDiff
      integer stage
      integer stride
      integer first
      real pwr,pwi,tr,ti
      real ir, ii, jr, ji
c     
c     bit reverse step
c     
      do i = 1,n
         j = brt(i)
         if (i .lt. j) then
            tr = a(1,j)
            ti = a(2,j)
            a(1,j) = a(1,i)
            a(2,j) = a(2,i)
            a(1,i) = tr
            a(2,i) = ti
         endif
      enddo
         
c     
c     butterfly computations
c     
      ijDiff = 1
      stride = 2
      sPowerOfW = ndv2
      do stage = 1,logn
c     Invariant: stride = 2 ** stage            
c     Invariant: ijDiff = 2 ** (stage - 1)
         
         first = 1
         do powerOfW = 1, ndv2, sPowerOfW
            pwr = W(1,powerOfW)
            pwi = W(2,powerOfW)
            
c     Invariant: pwr + sqrt(-1)*pwi = W**(powerOfW - 1)

            do i = first, n, stride
               j = i + ijDiff
               jr = a(1,j)
               ji = a(2,j)
               ir = a(1,i)
               ii = a(2,i)
               tr = jr*pwr - ji*pwi
               ti = jr*pwi + ji*pwr
               a(1,j) = ir - tr
               a(2,j) = ii - ti
               a(1,i) = ir + tr
               a(2,i) = ii + ti
            enddo

            first = first + 1
         enddo
            
         ijDiff = stride
         stride = stride * 2
         sPowerOfW = sPowerOfW / 2
      enddo
      return
      end

c
c     output routines
c

c     chkmat - check the results
c     
c     Since the results of the radar program are probabalistic, it is 
c     hard to write a data set independent routine to check the results.
c     However, it is easy to eyeball the results for channel 1 and check 
c     that they are correct. What you want to see in the det array is 
c     all zeros except near the doppler[k]'th element of the range0[k]'th
c     row of array det, k=0,..,NTARGETS, where you want to see a small 
c     number (3 or 4) of contiguous 1's.
c
      subroutine chkmat(det)
      include 'radar.h'
      integer det(NDOP,NRANGE,NCHAN)
      integer j,k
      
      print '(''***************************'')'
      print '(''Narrowband tracking radar'')'
      print '(i1,'' channels, '',i2,'' range cells, '')',NCHAN,NRANGE
      print '(i4,'' doppler cells, '',i2,'' targets'')',NDOP, NTARGETS
      print '(''***************************'')'
      print '(''Results for channel 1:'')'
      do j=1,NRANGE
         print '(40(I1))', (det(k,j,1),k=K0,K1)
      enddo
      print *, " "
      end

      subroutine prperf(tsecs,fsecs,dsecs,iters)
      real tsecs,fsecs,dsecs,secs
      integer tpercent,fpercent,dpercent,iters

      tsecs = tsecs/iters
      fsecs = fsecs/iters
      dsecs = dsecs/iters
      secs = tsecs+fsecs+dsecs
      
      tpercent = (tsecs/secs)*100
      fpercent = (fsecs/secs)*100
      dpercent = (dsecs/secs)*100
      
      print '(''Time for one set of returns:'')'
      print '(''tpose    : '',f9.6,'' ('',i2,''%)'')',tsecs,tpercent
      print '(''fft/mag  : '',f9.6,'' ('',i2,''%)'')',fsecs,fpercent
      print '(''detect   : '',f9.6,'' ('',i2,''%)'')',dsecs,dpercent
      print '(''total(s) : '',f9.6)',secs
      print *, " "
      end

      subroutine print_cmat(a,d,r,c)
      integer i,j,d,r,c
      real a(2,d,r,c)

      do j=1,r
         print '(8 (:''(''f6.2,f6.2'')'' 1x))',
     $        (a(1,i,j,1),a(2,i,j,1),i=1,d)
      enddo
      end

      subroutine print_cvec(v, n)
      integer n, i
      real v(2,n)

      do i=1,n
         print '(''(''f6.2,f6.2'')'')', v(1,i), v(2,i)
      enddo
      end

      subroutine print_ivec(v, n)
      integer n, v(n), i

      do i=1,n
         print *, v(i)
      enddo
      end

      subroutine print_rvec(v, n)
      integer n, i
      real v(n)

      do i=1,n
         print *, v(i)
      enddo
      end

c
c     seconds - returns elapsed seconds since last call to seconds().
c               the first time, returns elapsed time since program started. 
c
      real function seconds()
      real tarray(2),dtime
      seconds = dtime(tarray)
      seconds = tarray(1)
      end


